use chrono_tz::Europe;
use clickhouse_pool::{pool_manager::PoolManager, traits::Model};
use database::{
    trivial::{Trivial, TrivialStatus},
    trivial_round::TrivialRound,
};
use poise::serenity_prelude::Http;
use std::{
    collections::HashMap,
    fmt::{self, Display},
    sync::Arc,
};
use tokio::sync::RwLock;
use tokio_cron_scheduler::{job::JobLocked, Job, JobScheduler};
use tracing::{debug, error, info, info_span, instrument, Instrument};
use uuid::Uuid;

#[derive(Debug)]
pub enum StopScheduleJob {
    JobNotFound,
    RemoveFailed,
}

impl Display for StopScheduleJob {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(f, "{:?}", self)
    }
}

pub struct ScheduleJob {
    pub job_id: Arc<RwLock<HashMap<(u64, u64), Uuid>>>,
    pub scheduler: JobScheduler,
}

impl Clone for ScheduleJob {
    fn clone(&self) -> Self {
        ScheduleJob {
            job_id: self.job_id.clone(),
            scheduler: self.scheduler.clone(),
        }
    }
}

impl ScheduleJob {
    #[instrument(level = "info")]
    pub async fn start_cron_scheduler() -> Result<Self, ()> {
        let scheduler = JobScheduler::new().await;
        let mut future_self = match scheduler {
            Ok(scheduler) => ScheduleJob {
                job_id: Default::default(),
                scheduler,
            },
            Err(e) => {
                error!("Error starting cron scheduler: {:?}", e);
                return Err(());
            }
        };
        future_self.scheduler.set_shutdown_handler(Box::new(|| {
            Box::pin(async {
                info!("Cron scheduler stopped");
            })
        }));
        future_self.scheduler.shutdown_on_ctrl_c();
        match future_self.scheduler.start().await {
            Ok(_) => {
                info!("Cron scheduler started");
                Ok(future_self)
            }
            Err(e) => {
                error!("Error starting cron scheduler: {:?}", e);
                Err(())
            }
        }
    }

    #[instrument(skip(self, _http), level = "info")]
    pub async fn load_all_trivial_cron_job(&mut self, _http: &Http) -> Result<(), bool> {
        // Load all trivial jobs
        Ok(())
    }

    #[instrument(skip(self), level = "info")]
    pub async fn stop_cron_scheduler(&mut self) -> &mut Self {
        let job_id = self.job_id.write().await;

        for (server_id, channel_id) in job_id.keys() {
            match self
                .scheduler
                .remove(job_id.get(&(*server_id, *channel_id)).unwrap())
                .await
            {
                Ok(_) => {
                    info!("Cron job removed");
                }
                Err(e) => {
                    error!("Error removing cron job: {:?}", e);
                }
            }
        }

        drop(job_id);
        match self.scheduler.shutdown().await {
            Ok(_) => {
                info!("Cron scheduler stopped");
                self
            }
            Err(e) => {
                error!("Error stopping cron scheduler: {:?}", e);
                self
            }
        }
    }

    #[instrument(skip(self, http, manager), level = "info")]
    pub async fn add_day_end_trivial_cron_job(
        &mut self,
        guild_id: u64,
        channel_id: u64,
        cron_expression: String,
        manager: Arc<PoolManager>,
        http: &Http,
    ) -> Result<Uuid, ()> {
        let http = Arc::new(Http::new(http.token()));
        // Create a new job with the provided cron expression
        let job = match Job::new_cron_job_async_tz(
            cron_expression.as_str(),
            Europe::Paris,
            move |id, _l| {
                Box::pin(
                    {
                        let http = http.clone();
                        let manager = manager.clone();
                        Box::pin(async move {
                            info!(
                                id = id.to_string(),
                                guild_id = guild_id,
                                channel_id = channel_id,
                                "Cron job triggered for guild: {}, channel: {}",
                                guild_id,
                                channel_id
                            );
                            // Here you can add the logic to be executed when the cron job is triggered
                            // For example, sending a message to the channel
                            use poise::serenity_prelude::{ChannelId, CreateMessage};

                            let search_query = Trivial::build_select_query(
                                Some(&format!(
                                    "channel_id = {} and guild_id = {}",
                                    channel_id, guild_id
                                )),
                                None,
                                None,
                            );
                            let trivial_exists: Trivial = match manager
                              .execute_select_with_retry::<Trivial>(&search_query).await{
                                Ok(trivia) => {
                                    if trivia.is_empty() {
                                        info!("No trivia game found in this channel.");
                                        return;
                                    }
                                    if trivia.len() > 1 {
                                        error!("Multiple trivia games found in this channel, using the first one.");
                                    }
                                    trivia[0].clone()
                                }
                                Err(e) => {
                                    error!("Error fetching trivia game: {:?}", e);
                                    return;
                                }
                              };
                            debug!("Trivial game found: {:?}", trivial_exists);
                            if trivial_exists.status != TrivialStatus::Started {
                                info!("Trivia game is not started, skipping message.");
                                return;
                            }
                            let trivia_round_query = TrivialRound::build_select_query(
                                Some(&format!(
                                    "trivial_id = '{}' and finished = false",
                                    trivial_exists.id
                                )),
                                None,
                                None,
                            );
                            let trivia_round: TrivialRound = match manager
                                .execute_select_with_retry::<TrivialRound>(&trivia_round_query)
                                .await{
                                    Ok(rounds) => {
                                        if rounds.is_empty() {
                                            info!("No trivia rounds found for the current game.");
                                            return;
                                        }
                                        if rounds.len() > 1 {
                                            error!("Multiple trivia rounds found, using the first one.");
                                        }
                                        rounds[0].clone()
                                    },
                                    Err(e) => {
                                        error!("Error fetching trivia rounds: {:?}", e);
                                        return;
                                    }
                                };
                            debug!("Trivial round found: {:?}", trivia_round.id);
                            // TODO : Send a message to end the trivia round and that the bot is processing the results
                            // TODO : Add score to the user
                            // TODO : Update the trivial round to finished
                            // TODO : Congratulate the winner and delete te processing message
                            // TODO : Send message see you soon
                            if let Err(e) = ChannelId::new(channel_id)
                                .send_message(
                                    &http,
                                    CreateMessage::new().content("Cron job executed"),
                                )
                                .await
                            {
                                error!("Failed to send message: {:?}", e);
                            }
                        })
                    }
                    .instrument(info_span!(
                        "trivial_cron_job",
                        id = id.to_string(),
                        guild_id = guild_id,
                        channel_id = channel_id,
                    )),
                )
            },
        ) {
            Ok(job) => job,
            Err(e) => {
                error!("Error creating cron job: {:?}", e);
                return Err(());
            }
        };

        self.inject_job(job, (guild_id, channel_id)).await
    }

    #[instrument(skip(self, job), level = "info")]
    pub async fn inject_job(&mut self, job: JobLocked, id: (u64, u64)) -> Result<Uuid, ()> {
        match self.scheduler.add(job.clone()).await {
            Ok(job_uuid) => {
                info!(id = job_uuid.to_string(), "Cron job added successfully");
                {
                    let mut job_id = self.job_id.write().await;
                    job_id.insert(id, job_uuid);
                }
                Ok(job_uuid)
            }
            Err(e) => {
                error!("Error adding cron job: {:?}", e);
                return Err(());
            }
        }
    }

    #[instrument(skip(self), level = "info")]
    pub async fn stop_scheduled_job(
        &mut self,
        server_id: u64,
        channel_id: u64,
    ) -> Result<(), StopScheduleJob> {
        let remove_job = {
            let mut job_id = self.job_id.write().await;
            job_id.remove(&(server_id, channel_id))
        };
        match remove_job {
            Some(job_uid) => match self.scheduler.remove(&job_uid).await {
                Ok(_) => {
                    info!("Cron job removed");
                    Ok(())
                }
                Err(e) => {
                    error!("Error removing cron job: {:?}", e);
                    Err(StopScheduleJob::RemoveFailed)
                }
            },
            None => {
                error!("Cron job not found");
                Err(StopScheduleJob::JobNotFound)
            }
        }
    }
}
